@@ -1,7 +1,10 @@
 /* 
  * Copyright (C) 2006, 2007, 2008 OpenedHand Ltd.
+ * Copyright (C) 2009 Nokia Corporation, all rights reserved.
  *
  * Author: Jorn Baayen <jorn@openedhand.com>
+ *         Zeeshan Ali (Khattak) <zeeshanak@gnome.org>
+ *                               <zeeshan.ali@nokia.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -64,15 +67,16 @@ static const GSourceFuncs gssdp_socket_source_funcs = {
  * Return value: A new #GSSDPSocketSource
  **/
 GSSDPSocketSource *
-gssdp_socket_source_new (GSSDPSocketSourceType type)
+gssdp_socket_source_new (GSSDPSocketSourceType type,
+                         const char           *host_ip)
 {
         GSource *source;
         GSSDPSocketSource *socket_source;
-        struct sockaddr_in addr;
+        struct sockaddr_in bind_addr;
+        struct in_addr iface_addr;
         struct ip_mreq mreq;
         gboolean boolean = TRUE;
         guchar ttl = 4;
-        gushort port;
         int res;
 
         /* Create source */
@@ -110,6 +114,13 @@ gssdp_socket_source_new (GSSDPSocketSourceType type)
         if (res == -1)
                 goto error;
 
+        memset (&bind_addr, 0, sizeof (bind_addr));
+        bind_addr.sin_family = AF_INET;
+
+        res = inet_aton (host_ip, &iface_addr);
+        if (res == 0)
+                goto error;
+
         /* Set up additional things according to the type of socket desired */
         if (type == GSSDP_SOCKET_SOURCE_TYPE_MULTICAST) {
                 /* Allow multiple sockets to use the same PORT number */
@@ -121,9 +132,32 @@ gssdp_socket_source_new (GSSDPSocketSourceType type)
                 if (res == -1)
                         goto error;
 
+                /* Enable multicast loopback */
+                res = setsockopt (socket_source->poll_fd.fd,
+                                  IPPROTO_IP,
+                                  IP_MULTICAST_LOOP,
+                                  &boolean,
+                                  sizeof (boolean));
+                if (res == -1)
+                       goto error;
+
+                /* Set the interface */
+                res = setsockopt (socket_source->poll_fd.fd,
+                                  IPPROTO_IP,
+                                  IP_MULTICAST_IF,
+                                  &iface_addr,
+                                  sizeof (struct in_addr));
+                if (res == -1)
+                        goto error;
+
                 /* Subscribe to multicast channel */
-                mreq.imr_multiaddr.s_addr = inet_addr (SSDP_ADDR);
-                mreq.imr_interface.s_addr = htonl (INADDR_ANY);
+                res = inet_aton (SSDP_ADDR, &(mreq.imr_multiaddr));
+                if (res == 0)
+                        goto error;
+
+                memcpy (&(mreq.imr_interface),
+                        &iface_addr,
+                        sizeof (struct in_addr));
 
                 res = setsockopt (socket_source->poll_fd.fd,
                                   IPPROTO_IP,
@@ -133,20 +167,21 @@ gssdp_socket_source_new (GSSDPSocketSourceType type)
                 if (res == -1)
                         goto error;
 
-                port = SSDP_PORT;
-        } else
-                port = 0;
-       
-        /* Bind to requested port */
-        memset (&addr, 0, sizeof (addr));
-                
-        addr.sin_family      = AF_INET;
-        addr.sin_addr.s_addr = htonl (INADDR_ANY);
-        addr.sin_port        = htons (port);
+                bind_addr.sin_port = htons (SSDP_PORT);
+                res = inet_aton (SSDP_ADDR, &(bind_addr.sin_addr));
+                if (res == 0)
+                        goto error;
+        } else {
+                bind_addr.sin_port = 0;
+                memcpy (&(bind_addr.sin_addr),
+                        &iface_addr,
+                        sizeof (struct in_addr));
+        }
 
+        /* Bind to requested port and address */
         res = bind (socket_source->poll_fd.fd,
-                    (struct sockaddr *) &addr,
-                    sizeof (addr));
+                    (struct sockaddr *) &bind_addr,
+                    sizeof (bind_addr));
         if (res == -1)
                 goto error;
 
